-- Functions to abstract operations on list structures
-- Brad Olson, Jan 2003

-- NOTES: Future plans at the bottom of the file.
-- The tenor of these is to add some list and array processing abstractions
-- found in Lisp and APL dialects. None of these routines are necessary, they
-- just abstract some common patterns to try keep you from rewriting common code.


global function range( atom A, atom Z, atom Step)
--generates a list from a through z step n
  sequence result   result = {}
  for i = A to Z by Step do
    result &= i
  end for
  return result
end function

global function range0( integer n)
--numbers from 0..n-1
  return range( 0, n-1, 1)
end function

function benchmark( sequence Funs, sequence Args)
--Return timings of several functions over the same arguments
--e.g. benchmark( {"sin", "cos", tan"}, range(-1,1,0.01))
object X
atom T,N
  if length(Funs) then
    N = routine_id( Funs[1])
    T = -time()
    X = call_func( N, Args)
    T += time()
    printf( 1, "%10s  %10.4f\n", {Funs[1],T})
    return {Funs[1],T} & benchmark( Funs[2..length(Funs)], Args)
  else return {}
  end if
end function

global function every( integer Fun, object s)
--iterative version of every, loads faster (>40x) than a recursive version I tested
--call the function referred to by Fun (function id number) recursively on atoms of s
--return a hierarchical list of the same structure as s contain the results of Fun(atom)
--Fun needs to accept integers.  
--every() is not as fast as simple functions that automatically handle sublists, like
--InE below, and it's comparative performance gets worse as the nesting deepens
--But with more complex functions that start creating some large arguments, it's faster,
--like sinh (See the hyperbolic.e library).
--Incidentally, I think this would be one good candidate for implementing in C.
  integer len
  object x, result
  if atom(s) then   return call_func(Fun,{s})  end if
  len=length(s)
  result={}
  for i = 1 to len do
    x = s[i]
    if atom(x) then   result &= call_func(Fun,{x})
    else result &= { every( Fun,x) }
    end if
  end for
  return result
end function


---------------------------------------------------------------------------------------
-- IDEAS, Future plans  NONE OF THESE FUNCTIONS BELOW ARE IMPLEMENTED
-- The names and clarity of these abstractions need to be worked on.
---------------------------------------------------------------------------------------

--function map( integer Fun, object s)
--Apply Fun to each first-level item of s

--function mapZ(integer Fun, object s)
--An alias for every(), apply Fun to each atom of s down to the deepest level

--function mapN( integer Fun, object s, integer Rank)
--S is list of level >= rank. Slice s at the level Rank, apply Fun to each item, and
--reassemble in the format of s[Rank-1]'s structure
--This would be similar to n-dimensional slicing of arrays in APL
--Obviously it needs some more clarity first

--function reduce( integer Fun, object s)
--Reduce object s by applying Fun, which accepts two objects, successively between it's first-level items.
--e.g. if Fun = addition, return the sum of the list's items
--Might also want a reduceZ and reduceN


--function diff( integer Fun, object s)
--Returns a list of length(s)-1 items with Fun applied between each item and it's neighbor
